@@ -22,12 +22,14 @@ from . import angle
 
 import numpy as np
 import operator
-from scipy.interpolate import PchipInterpolator
 import math
 from scipy.spatial import ConvexHull
 from scipy.interpolate import griddata
 import matplotlib.pyplot as plt
 from functools import partial
+from scipy.optimize import curve_fit
+import uncertainties.unumpy as unp
+import uncertainties as unc
 
 # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 # ------------------------------------------------------------------------------
@@ -849,34 +851,53 @@ def disp1d(
 
     if smooth is True:
 
-        disp = PchipInterpolator(r[nonan], disp[nonan])(x[idxrange])
-        err = PchipInterpolator(r[nonan], err[nonan])(x[idxrange])
-        r = x[idxrange]
+        disp = disp[nonan]
+        err = err[nonan]
+        r = r[nonan]
 
         if polorder is None:
             pold = int(0.2 * position.good_bin(disp))
-            pole = int(0.2 * position.good_bin(err))
         else:
             pold = polorder
-            pole = polorder
 
         if logx is False:
-            poly_disp = np.polyfit(r, disp, pold)
-            poly_err = np.polyfit(r, err, pole)
 
-            disp = np.poly1d(poly_disp)(x[np.where(x < rmax)])
-            err = np.poly1d(poly_err)(x[np.where(x < rmax)])
-            r = x[np.where(x < rmax)]
+            poly_disp, cov_disp = np.polyfit(r, disp, pold, w=1 / err, cov=True)
+
+            # Do the interpolation for plotting:
+            t = x[idxrange]
+            # Matrix with rows 1, t, t**2, ...:
+            TT = np.vstack([t ** (pold - i) for i in range(pold + 1)]).T
+            yi = np.dot(
+                TT, poly_disp
+            )  # matrix multiplication calculates the polynomial values
+            C_yi = np.dot(TT, np.dot(cov_disp, TT.T))  # C_y = TT*C_z*TT.T
+            sig_yi = np.sqrt(np.diag(C_yi))  # Standard deviations are sqrt of diagonal
+
+            disp = yi
+            err = sig_yi
+            r = x[idxrange]
         else:
-            poly_disp = np.polyfit(np.log10(r), disp, pold)
-            poly_err = np.polyfit(np.log10(r), err, pole)
+            poly_disp, cov_disp = np.polyfit(
+                np.log10(r), disp, pold, w=1 / err, cov=True
+            )
 
-            disp = np.poly1d(poly_disp)(np.log10(x[idxrange]))
-            err = np.poly1d(poly_err)(np.log10(x[idxrange]))
+            # Do the interpolation for plotting:
+            t = np.log10(x[idxrange])
+            # Matrix with rows 1, t, t**2, ...:
+            TT = np.vstack([t ** (pold - i) for i in range(pold + 1)]).T
+            yi = np.dot(
+                TT, poly_disp
+            )  # matrix multiplication calculates the polynomial values
+            C_yi = np.dot(TT, np.dot(cov_disp, TT.T))  # C_y = TT*C_z*TT.T
+            sig_yi = np.sqrt(np.diag(C_yi))  # Standard deviations are sqrt of diagonal
+
+            disp = yi
+            err = sig_yi
             r = x[idxrange]
 
         if return_fits is True:
-            return r, disp, err, poly_disp, poly_err
+            return r, disp, err, poly_disp
 
     return r, disp, err
 
@@ -1266,6 +1287,475 @@ def bootstrap(array):
     return unc
 
 
+# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+# ------------------------------------------------------------------------------
+"Handles anisotropy"
+# ------------------------------------------------------------------------------
+
+
+def bgOM(r, ra, b0, bi):
+    """
+    Generalized (Osipkov1979; Merritt 1985) anisotropy profile.
+
+    Parameters
+    ----------
+    r : array_like, float
+        Distance from center of the system.
+    ra : float
+        Anisotropy radius.
+    b0 : float
+        Anisotropy at r = 0.
+    bi : float
+        Anisotropy at r -> Infinity.
+
+    Returns
+    -------
+    b : array_like, float
+        Anisotropy profile.
+
+    """
+
+    b = b0 + (bi - b0) / (1 + (ra / r) * (ra / r))
+
+    return b
+
+
+def bgTiret(r, ra, b0, bi):
+    """
+    Generalized Tiret et al. 2007 anisotropy profile.
+
+    Parameters
+    ----------
+    r : array_like, float
+        Distance from center of the system.
+    ra : float
+        Anisotropy radius.
+    b0 : float
+        Anisotropy at r = 0.
+    bi : float
+        Anisotropy at r -> Infinity.
+
+    Returns
+    -------
+    b : array_like, float
+        Anisotropy profile.
+
+    """
+
+    b = b0 + (bi - b0) / (1 + (ra / r))
+
+    return b
+
+
+def bCOM(r, ra, b0):
+    """
+    Anisotropy profile from (Cuddeford 1991; Osipkov 1979; Merritt 1985) inversion.
+
+    Parameters
+    ----------
+    r : array_like, float
+        Distance from center of the system.
+    ra : float
+        Anisotropy radius.
+    b0 : float
+        Anisotropy at r = 0.
+
+    Returns
+    -------
+    b : array_like, float
+        Anisotropy profile.
+
+    """
+
+    b = (b0 + (r / ra) * (r / ra)) / (1 + (r / ra) * (r / ra))
+
+    return b
+
+
+def get_anisotropy(
+    x,
+    y,
+    z,
+    vx,
+    vy,
+    vz,
+    method="moving",
+    nbin=30,
+    polorder=10,
+    logx=False,
+    bootp=False,
+    smooth=True,
+):
+    """
+    Computes the velocity anisotropy.
+
+     Parameters
+    ----------
+    x : array_like, float
+        x-axis.
+    y : array_like, float
+        y-axis.
+    z : array_like, float
+        z-axis.
+    vx : array_like, float
+        x-axis velocity.
+    vy : array_like, float
+        y-axis velocity.
+    vz : array_like, float
+        z-axis velocity.
+    method : string, optional
+        Method used to compute the dispersion.
+        The default is "moving".
+    nbin : int, optional
+        Number of bins/tracers used in the dispersion calculation.
+        The default is 30.
+    polorder : int, optional
+        Order of smoothing polynomial.
+        The default is 10.
+    logx : boolean, optional
+        If the x axis should be binned logarithmly.
+        The default is False.
+    bootp : boolean, optional
+        Use bootstrap to compute errors. The default is False.
+    smooth : boolean, optional
+        Smooth the data with a polynomial. The default is True.
+
+    Returns
+    -------
+    rr : array_like
+        Radius.
+    beta : array_like
+        Velocity anisotropy.
+    ebeta : array_like
+        Uncertainty on velocity anisotropy.
+
+    """
+
+    r, phi, theta, vr, vphi, vtheta = angle.cart_to_sph(x, y, z, vx, vy, vz)
+
+    L = sorted(zip(r, vr, vtheta, vphi), key=operator.itemgetter(0))
+    r, vr, vtheta, vphi = zip(*L)
+
+    r = np.asarray(r)
+    vr = np.asarray(vr)
+    vtheta = np.asarray(vtheta)
+    vphi = np.asarray(vphi)
+
+    rr, dr, err = dispersion(
+        r,
+        vr,
+        bins=method,
+        nbin=nbin,
+        polorder=polorder,
+        bootp=bootp,
+        logx=logx,
+        smooth=smooth,
+    )
+
+    rr, dt, ert = dispersion(
+        r,
+        vtheta,
+        bins=method,
+        nbin=nbin,
+        polorder=polorder,
+        bootp=bootp,
+        logx=logx,
+        smooth=smooth,
+    )
+
+    rr, dp, erp = dispersion(
+        r,
+        vphi,
+        bins=method,
+        nbin=nbin,
+        polorder=polorder,
+        bootp=bootp,
+        logx=logx,
+        smooth=smooth,
+    )
+
+    beta = 1 - (dt * dt + dp * dp) / (2 * dr * dr)
+
+    ebeta = (
+        0.5
+        * (dt * dt + dp * dp)
+        / (2 * dr * dr)
+        * np.sqrt(
+            2 * err * err / (dr * dr)
+            + 2 * (dt * dt * ert * ert + dp * dp * erp * erp) / (dt * dt + dp * dp) ** 2
+        )
+    )
+
+    return rr, beta, ebeta
+
+
+def get_beta(
+    x,
+    y,
+    z,
+    vx,
+    vy,
+    vz,
+    method="moving",
+    nbin=30,
+    polorder=10,
+    logx=False,
+    bootp=False,
+    smooth=True,
+    model="gOM",
+):
+    """
+    Computes the velocity anisotropy.
+
+     Parameters
+    ----------
+    x : array_like, float
+        x-axis.
+    y : array_like, float
+        y-axis.
+    z : array_like, float
+        z-axis.
+    vx : array_like, float
+        x-axis velocity.
+    vy : array_like, float
+        y-axis velocity.
+    vz : array_like, float
+        z-axis velocity.
+    method : string, optional
+        Method used to compute the dispersion.
+        The default is "moving".
+    nbin : int, optional
+        Number of bins/tracers used in the dispersion calculation.
+        The default is 30.
+    polorder : int, optional
+        Order of smoothing polynomial.
+        The default is 10.
+    logx : boolean, optional
+        If the x axis should be binned logarithmly.
+        The default is False.
+    bootp : boolean, optional
+        Use bootstrap to compute errors. The default is False.
+    smooth : boolean, optional
+        Smooth the data with a polynomial. The default is True.
+    model : string, optional
+        Model used to fit (or not fit) the data. The default is "gOM".
+
+    Raises
+    ------
+    ValueError
+        Anisotropy model is not one of the following:
+            - 'gOM'
+            - 'gTiret'
+            - 'gCOM'
+            - 'polynomial'
+            - 'discrete'
+
+    Returns
+    -------
+    rr : array_like
+        Radius.
+    beta : array_like
+        Velocity anisotropy.
+    ebeta : array_like
+        Uncertainty on velocity anisotropy.
+
+    """
+
+    if model not in ["gOM", "gTiret", "gCOM", "polynomial", "discrete"]:
+        raise ValueError("Does not recognize surface density model.")
+
+    if model == "discrete":
+        smooth = False
+
+    if smooth is False or model == "polynomial":
+        rr, beta, ebeta = get_anisotropy(
+            x,
+            y,
+            z,
+            vx,
+            vy,
+            vz,
+            method=method,
+            nbin=nbin,
+            polorder=polorder,
+            bootp=bootp,
+            logx=logx,
+            smooth=smooth,
+        )
+    else:
+        rr, beta, ebeta = get_anisotropy(
+            x,
+            y,
+            z,
+            vx,
+            vy,
+            vz,
+            method=method,
+            nbin=nbin,
+            polorder=polorder,
+            bootp=bootp,
+            logx=logx,
+            smooth=False,
+        )
+
+        r, phi, theta, vr, vphi, vtheta = angle.cart_to_sph(x, y, z, vx, vy, vz)
+        r = np.sort(r)
+
+        nonan1 = np.logical_not(np.isnan(beta))
+        nonan2 = np.logical_not(np.isnan(ebeta))
+        nonan = nonan1 * nonan2
+
+        rr = rr[nonan]
+        beta = beta[nonan]
+        ebeta = ebeta[nonan]
+
+        # https://stackoverflow.com/questions/24633664/confidence-interval-for-exponential-curve-fit/26042460#26042460
+
+        if model == "gOM":
+
+            popt, pcov = curve_fit(
+                bgOM,
+                rr,
+                beta,
+                sigma=ebeta,
+                p0=[np.nanmean(r), 0, 0],
+                absolute_sigma=False,
+                bounds=([0, -np.inf, -np.inf], [np.inf, 1.0, 1.0]),
+            )
+
+            b_params = unc.correlated_values(popt, pcov)
+
+            px = r
+            py = bgOM(px, *b_params)
+
+        if model == "gTiret":
+
+            popt, pcov = curve_fit(
+                bgTiret,
+                rr,
+                beta,
+                sigma=ebeta,
+                p0=[np.nanmean(r), 0, 0],
+                absolute_sigma=False,
+                bounds=([0, -np.inf, -np.inf], [np.inf, 1.0, 1.0]),
+            )
+
+            b_params = unc.correlated_values(popt, pcov)
+
+            px = r
+            py = bgTiret(px, *b_params)
+
+        if model == "gCOM":
+
+            popt, pcov = curve_fit(
+                bCOM,
+                rr,
+                beta,
+                sigma=ebeta,
+                p0=[np.nanmean(r), 0],
+                absolute_sigma=False,
+                bounds=([0, -np.inf], [np.inf, 1.0]),
+            )
+
+            b_params = unc.correlated_values(popt, pcov)
+
+            px = r
+            py = bCOM(px, *b_params)
+
+        rr = px
+        beta = unp.nominal_values(py)
+        ebeta = unp.std_devs(py)
+
+    return rr, beta, ebeta
+
+
+def get_betasym(
+    x,
+    y,
+    z,
+    vx,
+    vy,
+    vz,
+    method="moving",
+    nbin=30,
+    polorder=10,
+    logx=False,
+    bootp=False,
+    smooth=True,
+    model="gOM",
+):
+    """
+    Computes the symmetric velocity anisotropy, defined as:
+        beta_sym(r) = beta(r) / (1 - beta(r)/2)
+
+    Parameters
+    ----------
+    x : array_like, float
+        x-axis.
+    y : array_like, float
+        y-axis.
+    z : array_like, float
+        z-axis.
+    vx : array_like, float
+        x-axis velocity.
+    vy : array_like, float
+        y-axis velocity.
+    vz : array_like, float
+        z-axis velocity.
+    method : string, optional
+        Method used to compute the dispersion.
+        The default is "moving".
+    nbin : int, optional
+        Number of bins/tracers used in the dispersion calculation.
+        The default is 30.
+    polorder : int, optional
+        Order of smoothing polynomial.
+        The default is 10.
+    logx : boolean, optional
+        If the x axis should be binned logarithmly.
+        The default is False.
+    bootp : boolean, optional
+        Use bootstrap to compute errors. The default is False.
+    smooth : boolean, optional
+        Smooth the data with a polynomial. The default is True.
+    model : string, optional
+            Model used to fit (or not fit) the data. The default is "gOM".
+
+    Returns
+    -------
+    rr : array_like
+        Radius.
+    beta : array_like
+        Symmetric velocity anisotropy.
+    ebeta : array_like
+        Uncertainty on the symmetric velocity anisotropy.
+
+    """
+    r, beta, ebeta = get_beta(
+        x,
+        y,
+        z,
+        vx,
+        vy,
+        vz,
+        method=method,
+        nbin=nbin,
+        polorder=polorder,
+        bootp=bootp,
+        logx=logx,
+        smooth=smooth,
+        model=model,
+    )
+
+    beta_sym = beta / (1 - beta * 0.5)
+
+    ebeta_sym = np.abs(beta_sym * ebeta) * np.sqrt(
+        1 / (beta * beta) + 0.25 / ((1 - beta * 0.5) * (1 - beta * 0.5))
+    )
+
+    return r, beta_sym, ebeta_sym
+
+
 # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 # ------------------------------------------------------------------------------
 "Plotting functions"
